@@ -61,6 +61,7 @@ enum
 static GObject *xfce_titled_dialog_constructor    (GType                   type,
                                                    guint                   n_construct_params,
                                                    GObjectConstructParam  *construct_params);
+static void     xfce_titled_dialog_constructed    (GObject                *object);
 static void     xfce_titled_dialog_finalize       (GObject                *object);
 static void     xfce_titled_dialog_get_property   (GObject                *object,
                                                    guint                   prop_id,
@@ -72,18 +73,18 @@ static void     xfce_titled_dialog_set_property   (GObject                *objec
                                                    GParamSpec             *pspec);
 static void     xfce_titled_dialog_close          (GtkDialog              *dialog);
 static void     xfce_titled_dialog_update_window  (XfceTitledDialog       *titled_dialog);
-static void     xfce_titled_dialog_update_icon    (XfceTitledDialog       *titled_dialog);
+static void     xfce_titled_dialog_update_layout  (GObject                *settings,
+                                                   GParamSpec             *pspec,
+                                                   XfceTitledDialog       *titled_dialog);
 
 
 
 struct _XfceTitledDialogPrivate
 {
   GtkWidget *headerbar;
-  GtkWidget *icon;
   GtkWidget *action_area;
   GtkWidget *subtitle_label;
   GtkWidget *subtitle_separator;
-  GdkPixbuf *pixbuf;
   gchar     *subtitle;
   gboolean   use_header;
 };
@@ -110,6 +111,7 @@ xfce_titled_dialog_class_init (XfceTitledDialogClass *klass)
 
   gobject_class = G_OBJECT_CLASS (klass);
   gobject_class->constructor = xfce_titled_dialog_constructor;
+  gobject_class->constructed = xfce_titled_dialog_constructed;
   gobject_class->get_property = xfce_titled_dialog_get_property;
   gobject_class->set_property = xfce_titled_dialog_set_property;
   gobject_class->finalize = xfce_titled_dialog_finalize;
@@ -166,6 +168,10 @@ xfce_titled_dialog_init (XfceTitledDialog *titled_dialog)
   settings = gtk_settings_get_default ();
   g_object_get (settings, "gtk-dialogs-use-header", &titled_dialog->priv->use_header, NULL);
 
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+  titled_dialog->priv->action_area = gtk_dialog_get_action_area (GTK_DIALOG (titled_dialog));
+G_GNUC_END_IGNORE_DEPRECATIONS
+
   if (titled_dialog->priv->use_header)
     {
       g_object_set (G_OBJECT (titled_dialog), "use-header-bar", TRUE, NULL);
@@ -177,19 +183,14 @@ xfce_titled_dialog_init (XfceTitledDialog *titled_dialog)
       /* Don't reserve vertical space for subtitles */
       gtk_header_bar_set_has_subtitle (GTK_HEADER_BAR (titled_dialog->priv->headerbar), FALSE);
 
-      /* Pack the window icon into the headerbar */
-      titled_dialog->priv->icon = gtk_image_new ();
-      gtk_header_bar_pack_start (GTK_HEADER_BAR (titled_dialog->priv->headerbar), titled_dialog->priv->icon);
-      gtk_header_bar_set_show_close_button (GTK_HEADER_BAR (titled_dialog->priv->headerbar), TRUE);
-      gtk_widget_show (titled_dialog->priv->icon);
-      titled_dialog->priv->pixbuf = NULL;
-
       /* Adjust window buttons and window placement */
+      gtk_header_bar_set_show_close_button (GTK_HEADER_BAR (titled_dialog->priv->headerbar), TRUE);
       g_signal_connect (G_OBJECT (titled_dialog), "notify::window", G_CALLBACK (xfce_titled_dialog_update_window), NULL);
 
-      /* Make sure to update the icon whenever one of the relevant window properties changes */
-      g_signal_connect (G_OBJECT (titled_dialog), "notify::icon", G_CALLBACK (xfce_titled_dialog_update_icon), NULL);
-      g_signal_connect (G_OBJECT (titled_dialog), "notify::icon-name", G_CALLBACK (xfce_titled_dialog_update_icon), NULL);
+      /* Add icon to the layout if needed */
+      xfce_titled_dialog_update_layout (G_OBJECT (settings), NULL, titled_dialog);
+      g_signal_connect_object (settings, "notify::gtk-decoration-layout",
+                               G_CALLBACK (xfce_titled_dialog_update_layout), titled_dialog, 0);
     }
   else
     {
@@ -223,6 +224,57 @@ xfce_titled_dialog_init (XfceTitledDialog *titled_dialog)
 
 
 
+static void
+xfce_titled_dialog_constructed (GObject *object)
+{
+  XfceTitledDialog *titled_dialog = XFCE_TITLED_DIALOG (object);
+  GList *children = NULL;
+
+  /* remove and save all buttons from action area */
+  if (titled_dialog->priv->use_header)
+    {
+      children = gtk_container_get_children (GTK_CONTAINER (titled_dialog->priv->action_area));
+
+      for (GList *l = children; l != NULL; l = l->next)
+        {
+          g_object_ref (l->data);
+          gtk_container_remove (GTK_CONTAINER (titled_dialog->priv->action_area), l->data);
+        }
+    }
+
+  G_OBJECT_CLASS (xfce_titled_dialog_parent_class)->constructed (object);
+
+  if (titled_dialog->priv->use_header)
+    {
+      /* undo some unwanted GTK actions in constructed() */
+      GtkWidget *action_box = gtk_widget_get_parent (titled_dialog->priv->action_area);
+      gtk_widget_set_no_show_all (action_box, FALSE);
+      gtk_widget_show (action_box);
+      g_signal_handlers_disconnect_matched (titled_dialog->priv->action_area,
+                                            G_SIGNAL_MATCH_ID | G_SIGNAL_MATCH_DATA,
+                                            g_signal_lookup ("add", GTK_TYPE_CONTAINER),
+                                            0, NULL, NULL, titled_dialog);
+
+      /* putting back buttons to action area */
+      for (GList *l = children; l != NULL; l = l->next)
+        {
+          ResponseData *rd = g_object_get_data (l->data, "gtk-dialog-response-data");
+          gint response_id = rd != NULL ? rd->response_id : GTK_RESPONSE_NONE;
+
+          gtk_container_add (GTK_CONTAINER (titled_dialog->priv->action_area), l->data);
+          g_object_unref (l->data);
+
+          /* always add help button as secondary */
+          if (response_id == GTK_RESPONSE_HELP)
+            gtk_button_box_set_child_secondary (GTK_BUTTON_BOX (titled_dialog->priv->action_area), l->data, TRUE);
+        }
+
+      g_list_free (children);
+    }
+}
+
+
+
 static void
 xfce_titled_dialog_finalize (GObject *object)
 {
@@ -231,13 +283,6 @@ xfce_titled_dialog_finalize (GObject *object)
   /* release the subtitle */
   g_free (titled_dialog->priv->subtitle);
 
-  /* release the pixbuf */
-  if (titled_dialog->priv->pixbuf)
-    {
-      g_object_unref (titled_dialog->priv->pixbuf);
-      titled_dialog->priv->pixbuf = NULL;
-    }
-
   (*G_OBJECT_CLASS (xfce_titled_dialog_parent_class)->finalize) (object);
 }
 
@@ -323,27 +368,29 @@ xfce_titled_dialog_update_window (XfceTitledDialog *titled_dialog)
 
 
 static void
-xfce_titled_dialog_update_icon (XfceTitledDialog *titled_dialog)
+xfce_titled_dialog_update_layout (GObject *settings,
+                                  GParamSpec *pspec,
+                                  XfceTitledDialog *titled_dialog)
 {
-  const gchar *icon_name;
+  gchar *layout;
 
-  icon_name = gtk_window_get_icon_name (GTK_WINDOW (titled_dialog));
+  g_object_get (settings, "gtk-decoration-layout", &layout, NULL);
 
-  if (icon_name)
+  /* add icon on the far left of the header bar if not already present */
+  if (g_strstr_len (layout, -1, "icon") == NULL)
     {
-      gtk_image_set_from_icon_name (GTK_IMAGE (titled_dialog->priv->icon), icon_name, GTK_ICON_SIZE_LARGE_TOOLBAR);
-      gtk_image_set_pixel_size (GTK_IMAGE (titled_dialog->priv->icon), 24);
-    }
-  else
-    {
-      if (titled_dialog->priv->pixbuf)
-          g_object_unref (titled_dialog->priv->pixbuf);
-      titled_dialog->priv->pixbuf = gtk_window_get_icon (GTK_WINDOW (titled_dialog));
-      gtk_image_set_from_pixbuf (GTK_IMAGE (titled_dialog->priv->icon), titled_dialog->priv->pixbuf);
+      const gchar *separator = (g_strstr_len (layout, -1, ":") == NULL) ? ":" : ",";
+      gchar *layout_with_icon = g_strconcat ("icon", separator, layout, NULL);
+      g_free (layout);
+      layout = layout_with_icon;
     }
+
+  gtk_header_bar_set_decoration_layout (GTK_HEADER_BAR (titled_dialog->priv->headerbar), layout);
+  g_free (layout);
 }
 
 
+
 /* Borrowed from gtkdialog.c */
 static void
 response_data_free (gpointer data)
@@ -418,28 +465,6 @@ add_response_data (GtkDialog *dialog,
 
 
 
-/* Repack all the buttons that would normally end up in the headerbar to the action area */
-static void
-xfce_titled_dialog_repack_dialog (GtkWidget *action_area,
-                                  GtkWidget *headerbar,
-                                  GtkWidget *button,
-                                  gint       response_id)
-{
-  g_return_if_fail (GTK_IS_WIDGET (action_area));
-  g_return_if_fail (GTK_IS_WIDGET (headerbar));
-  g_return_if_fail (GTK_IS_WIDGET (button));
-
-  g_object_ref (G_OBJECT (button));
-  gtk_container_remove (GTK_CONTAINER (headerbar), button);
-  gtk_container_add (GTK_CONTAINER (action_area), button);
-  g_object_unref (G_OBJECT (button));
-  /* always add help buttons as secondary */
-  if (response_id == GTK_RESPONSE_HELP)
-    gtk_button_box_set_child_secondary (GTK_BUTTON_BOX (action_area), button, TRUE);
-}
-
-
-
 /**
  * xfce_titled_dialog_new:
  *
@@ -457,10 +482,10 @@ xfce_titled_dialog_new (void)
 
 /**
  * xfce_titled_dialog_new_with_buttons:
- * @title             : (allow-none): title of the dialog, or %NULL.
- * @parent            : (allow-none): transient parent window of the dialog, or %NULL.
+ * @title             : (nullable): title of the dialog, or %NULL.
+ * @parent            : (nullable): transient parent window of the dialog, or %NULL.
  * @flags             : from #GtkDialogFlags.
- * @first_button_text : (allow-none): stock ID or text to go in first, or %NULL.
+ * @first_button_text : (nullable): stock ID or text to go in first, or %NULL.
  * @...               : response ID for the first button, then additional buttons, ending with %NULL.
  *
  * See the documentation of gtk_dialog_new_with_buttons() for details about the
@@ -480,8 +505,6 @@ xfce_titled_dialog_new_with_buttons (const gchar    *title,
   const gchar *button_text;
   GtkWidget   *dialog;
   GtkWidget   *button;
-  GtkWidget   *headerbar;
-  GtkWidget   *action_area;
   va_list      args;
   gint         response_id;
 
@@ -496,11 +519,6 @@ xfce_titled_dialog_new_with_buttons (const gchar    *title,
   if (G_LIKELY (parent != NULL))
     gtk_window_set_transient_for (GTK_WINDOW (dialog), parent);
 
-G_GNUC_BEGIN_IGNORE_DEPRECATIONS
-  action_area = gtk_dialog_get_action_area (GTK_DIALOG (dialog));
-G_GNUC_END_IGNORE_DEPRECATIONS
-  headerbar = gtk_dialog_get_header_bar (GTK_DIALOG (dialog));
-
   /* add all additional buttons */
   va_start (args, first_button_text);
   for (button_text = first_button_text; button_text != NULL; )
@@ -509,9 +527,7 @@ G_GNUC_END_IGNORE_DEPRECATIONS
 G_GNUC_BEGIN_IGNORE_DEPRECATIONS
       button = gtk_button_new_from_stock (button_text);
 G_GNUC_END_IGNORE_DEPRECATIONS
-      gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, response_id);
-      if (XFCE_TITLED_DIALOG (dialog)->priv->use_header)
-        xfce_titled_dialog_repack_dialog (action_area, headerbar, button, response_id);
+      xfce_titled_dialog_add_action_widget (XFCE_TITLED_DIALOG (dialog), button, response_id);
       button_text = va_arg (args, const gchar *);
     }
   va_end (args);
@@ -523,11 +539,11 @@ G_GNUC_END_IGNORE_DEPRECATIONS
 
 /**
  * xfce_titled_dialog_new_with_mixed_buttons:
- * @title                  :  (allow-none):title of the dialog, or %NULL.
- * @parent                 : (allow-none): transient parent window of the dialog, or %NULL.
+ * @title                  : (nullable): title of the dialog, or %NULL.
+ * @parent                 : (nullable): transient parent window of the dialog, or %NULL.
  * @flags                  : from #GtkDialogFlags.
  * @first_button_icon_name : icon name to go in first, or "" for no icon.
- * @first_button_text      : (allow-none): text to go in first, or %NULL.
+ * @first_button_text      : (nullable): text to go in first, or %NULL.
  * @...                    : response ID for the first button, then additional buttons, ending with %NULL.
  *
  * Creates an #XfceTitledDialog using xfce_gtk_button_new_mixed. This allows
@@ -549,8 +565,6 @@ xfce_titled_dialog_new_with_mixed_buttons (const gchar    *title,
   const gchar *icon_name;
   const gchar *button_text;
   GtkWidget   *dialog;
-  GtkWidget   *headerbar;
-  GtkWidget   *action_area;
   va_list      args;
   gint         response_id;
 
@@ -565,11 +579,6 @@ xfce_titled_dialog_new_with_mixed_buttons (const gchar    *title,
   if (G_LIKELY (parent != NULL))
     gtk_window_set_transient_for (GTK_WINDOW (dialog), parent);
 
-G_GNUC_BEGIN_IGNORE_DEPRECATIONS
-  action_area = gtk_dialog_get_action_area (GTK_DIALOG (dialog));
-G_GNUC_END_IGNORE_DEPRECATIONS
-  headerbar = gtk_dialog_get_header_bar (GTK_DIALOG (dialog));
-
   /* add all additional buttons */
   icon_name = first_button_icon_name;
   button_text = first_button_text;
@@ -586,9 +595,7 @@ G_GNUC_END_IGNORE_DEPRECATIONS
       button = xfce_gtk_button_new_mixed (icon_name, button_text);
       gtk_widget_set_can_default (button, TRUE);
 
-      gtk_dialog_add_action_widget (GTK_DIALOG (dialog), button, response_id);
-      if (XFCE_TITLED_DIALOG (dialog)->priv->use_header)
-        xfce_titled_dialog_repack_dialog (action_area, headerbar, button, response_id);
+      xfce_titled_dialog_add_action_widget (XFCE_TITLED_DIALOG (dialog), button, response_id);
       gtk_widget_show (button);
 
       /* this is to pickup for the next button.
@@ -602,10 +609,6 @@ G_GNUC_END_IGNORE_DEPRECATIONS
     }
   va_end (args);
 
-  /* since we just removed all buttons from the headerbar we have to show the close button again explicitly */
-  if (XFCE_TITLED_DIALOG (dialog)->priv->use_header)
-    gtk_header_bar_set_show_close_button (GTK_HEADER_BAR (headerbar), TRUE);
-
   return dialog;
 }
 
@@ -615,33 +618,15 @@ G_GNUC_END_IGNORE_DEPRECATIONS
  * xfce_titled_dialog_create_action_area:
  * @titled_dialog : a #XfceTitledDialog.
  *
- * This function creates a custom action area (of type #GtkButtonBox) and has to
- * be used in combination with #xfce_titled_dialog_add_action_widget.
- *
- * When using the XfceTitledDialogClass directly to create dialogs this function is
- * useful to keep action widgets out of the #GtkHeaderBar in which they would
- * normally end up by calling #gtk_dialog_add_action_widget.
+ * This function is a no-op since 4.19.3.
  *
  * Since: 4.16
  *
+ * Deprecated: 4.19.3
  **/
 void
 xfce_titled_dialog_create_action_area (XfceTitledDialog *titled_dialog)
 {
-  GtkWidget *dialog_vbox;
-
-  g_return_if_fail (XFCE_IS_TITLED_DIALOG (titled_dialog));
-
-  /* Create new buttonbox to act as custom action area for the dialog */
-  titled_dialog->priv->action_area = gtk_button_box_new (GTK_ORIENTATION_HORIZONTAL);
-  gtk_button_box_set_layout (GTK_BUTTON_BOX (titled_dialog->priv->action_area), GTK_BUTTONBOX_END);
-  gtk_box_set_spacing (GTK_BOX (titled_dialog->priv->action_area), 6);
-
-  /* Pack the buttonbox in the dialog's main vbox */
-  dialog_vbox = gtk_bin_get_child (GTK_BIN (titled_dialog));
-  gtk_box_pack_end (GTK_BOX (dialog_vbox), titled_dialog->priv->action_area, FALSE, TRUE, 0);
-  gtk_box_reorder_child (GTK_BOX (dialog_vbox), titled_dialog->priv->action_area, 0);
-  gtk_widget_show (titled_dialog->priv->action_area);
 }
 
 
@@ -652,8 +637,7 @@ xfce_titled_dialog_create_action_area (XfceTitledDialog *titled_dialog)
  * @button_text   : text of button.
  * @response_id   : response ID for @child.
  *
- * This function is a replacement for #gtk_dialog_add_button and assumes that
- * you have called #xfce_titled_dialog_create_action_area before.
+ * This function is a replacement for #gtk_dialog_add_button.
  *
  * Buttons with #GTK_RESPONSE_HELP will be added to the secondary group of children
  * (see #gtk_button_box_set_child_secondary for reference).
@@ -690,8 +674,7 @@ xfce_titled_dialog_add_button (XfceTitledDialog *titled_dialog,
  * @child         : an activatable widget.
  * @response_id   : response ID for @child.
  *
- * This function is a replacement for #gtk_dialog_add_action_widget and assumes that
- * you have called #xfce_titled_dialog_create_action_area before.
+ * This function is a replacement for #gtk_dialog_add_action_widget.
  *
  * Children with #GTK_RESPONSE_HELP will be added to the secondary group of children
  * (see #gtk_button_box_set_child_secondary for reference).
@@ -738,17 +721,12 @@ void
 xfce_titled_dialog_set_default_response (XfceTitledDialog *titled_dialog,
                                          gint              response_id)
 {
-  GtkWidget *action_area;
   GList     *children;
   GList     *tmp_list;
 
   g_return_if_fail (XFCE_IS_TITLED_DIALOG (titled_dialog));
 
-G_GNUC_BEGIN_IGNORE_DEPRECATIONS
-  action_area = gtk_dialog_get_action_area (GTK_DIALOG (titled_dialog));
-G_GNUC_END_IGNORE_DEPRECATIONS
-
-  children = gtk_container_get_children (GTK_CONTAINER (action_area));
+  children = gtk_container_get_children (GTK_CONTAINER (titled_dialog->priv->action_area));
   tmp_list = children;
   while (tmp_list != NULL)
     {
@@ -756,7 +734,10 @@ G_GNUC_END_IGNORE_DEPRECATIONS
       ResponseData *rd = get_response_data (widget, FALSE);
 
       if (rd && rd->response_id == response_id)
-        gtk_window_set_default (GTK_WINDOW (titled_dialog), widget);
+        {
+          gtk_widget_set_can_default (widget, TRUE);
+          gtk_window_set_default (GTK_WINDOW (titled_dialog), widget);
+        }
 
       tmp_list = tmp_list->next;
     }
@@ -788,7 +769,7 @@ xfce_titled_dialog_get_subtitle (XfceTitledDialog *titled_dialog)
 /**
  * xfce_titled_dialog_set_subtitle:
  * @titled_dialog : a #XfceTitledDialog.
- * @subtitle      : the new subtitle for the @titled_dialog, or %NULL.
+ * @subtitle: (nullable): the new subtitle for the @titled_dialog, or %NULL.
  *
  * Sets the subtitle displayed by @titled_dialog to @subtitle; if
  * @subtitle is %NULL no subtitle will be displayed by the @titled_dialog.
